import io
import re
import time
import socket
import logging
import contextlib
from collections import namedtuple
from typing import Union, Tuple, NamedTuple, Pattern, Callable, NewType

import paramiko

from . import _ssh, exceptions

logger = logging.getLogger()

_match_data = namedtuple("match_data", ["string", "groups"])
_command_return = namedtuple("command_return", ["output", "match_info"])
CommandReturn = NewType("CommandReturn", _command_return)


@contextlib.contextmanager
def _connect(
    address: str,
    port: Union[int, str],
    username: Union[str, None] = None,
    password: Union[str, None] = None,
    private_key: Union[str, None] = None,
    passphrase: Union[str, None] = None,
    allow_agent: bool = False,
    timeout: int = 10,
    base_prompt: str = None,
    shell_timeout: int = 5,
    missing_host_key_policy: paramiko.MissingHostKeyPolicy = paramiko.AutoAddPolicy(),
):

    prompt_fn = _regex_search(base_prompt)

    client = _ssh._get_client(
        address, port, username, password, private_key, passphrase, allow_agent, timeout, missing_host_key_policy
    )

    shell = _ssh._get_shell(client, shell_timeout)

    def _reader():
        """
        Receive data from the shell either until the prompt is
        found, or the shell timeout is reached.
        """
        buffer = io.BytesIO()

        while True:
            try:
                data = shell.recv(8192)
                if not data:
                    break
                buffer.write(data)
                yield buffer.getvalue()
            except socket.timeout as e:
                logger.error(f"no data received for {shell_timeout} seconds, assuming buffer is empty")
                raise e

    def _next(prompt_fn):
        """
        Read from the buffer until the prompt is found. Return the
        data up to the start of the prompt.
        """
        try:
            data = None
            for data in _reader():
                match = prompt_fn(data)
                if match:
                    output_start, match_start, match_info = match
                    return (data[output_start:match_start], match_info)
            else:
                # If the data runs out before either the prompt
                # is matched or the shell times out, force a socket
                # timeout to trigger the exception.
                raise socket.timeout("shell buffer was exhausted before timing out")
        except socket.timeout:
            raise exceptions.PatternNotFound(f"Could not find expected prompt {prompt_fn._pattern} in response {data}.")

    # consume and throw away the intial banner
    _banner = _next(prompt_fn)  # noqa: F841

    def _run(command, alt_prompt_fn=None):
        """ Send a command to the device and run prompt_fn with
        the response. If the shell isn't ready to receive a command
        within 5 seconds, raise a OperationTimeout.
        """

        def _wait():
            counter = 0
            while counter <= shell_timeout:
                if shell.send_ready():
                    return
                time.sleep(1)
                counter += 1
            raise exceptions.OperationTimeout(f"Shell not send ready after {shell_timeout} seconds.")

        _wait()
        shell.send(command + "\n")
        return _next(alt_prompt_fn if alt_prompt_fn else prompt_fn)

    try:
        yield _run
    finally:
        shell.close()
        client.close()


def _clean_response(*, response: bytes, command: str) -> str:
    """ Decode the bytes response to utf-8 and replace
    any echoed back commands. This function takes keyword-only
    arguments for readability.
    """
    logger.debug(f"cleaning response '{response}'")
    return response.decode("utf-8").replace(command, "").replace("\r", "").strip()


def _regex_search(pattern: Union[str, bytes, Pattern]) -> Callable:
    """ Attempt to match the supplied regex against the data
    returned by the device.

    The values returned by this function:

    - Start index of the command output (always 0)
    - Start index of the matched prompt
    - Either a match group dict for named capturing groups,
    or the matched prompt as a string.
    """
    if isinstance(pattern, str):
        pattern = bytes(pattern, "utf-8")

    try:
        pattern = re.compile(pattern, re.MULTILINE)
    except re.error as e:
        raise exceptions.InvalidRegex(f"provided pattern {pattern} is not a valid regex - {str(e)}.")

    def fn(data: bytes) -> Union[Tuple[int, int, NamedTuple], None]:
        matchdata = re.search(pattern, data)  # type: ignore
        if matchdata:
            group_dict = matchdata.groupdict()
            return (  # type: ignore
                0,
                matchdata.start(),
                _match_data(
                    string=data[matchdata.start() : matchdata.end()].decode("utf-8"),
                    groups=group_dict if group_dict else None,
                ),
            )
        return None

    setattr(fn, "_pattern", pattern.pattern)  #  For exception messages.
    return fn


def _string_search(pattern: Union[str, bytes]) -> Callable:
    """ Look for the supplied pattern in the data returned
    by the device.

    The values returned by this function:

    - Start index of the command output (always 0)
    - Start index of the matched prompt
    - String matched by the prompt.
    """
    if isinstance(pattern, str):
        pattern = bytes(pattern, "utf-8")

    def fn(data: bytes) -> Union[Tuple[int, int, NamedTuple], None]:
        if pattern in data:  # type: ignore
            start_index = data.find(pattern)  # type: ignore
            return (0, start_index, _match_data(string=data[start_index:].decode("utf-8"), groups=None))  # type: ignore
        return None

    setattr(fn, "_pattern", pattern)  #  For exception messages.
    return fn
