"""
Config

This module contains actions for copying & writing configuration on a Cisco IOS v15 switch.

See the [Overview](overview.md) page for information on how actions in this plugin work.
"""
import logging

from typing import Optional

from ssh_plugin_driver.client import run_list

from .records import CiscoSwitch
from ._generic_functions import optional_progress_bar, with_enable_mode, check_command_responses, session

logger = logging.getLogger(__name__)


@session
def copy_running_to_startup(
    cisco_switch: CiscoSwitch,
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 180,
    connection_timeout: Optional[int] = 10,
) -> None:
    """Copy the running configuration to startup-config.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __enable_password:__ an [enable password](overview.md#enable-mode) if required

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `180`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.config.copy_running_to_startup:
        progress_bar: true
      load:
        cisco_switch: cisco_switch
    ```
    """
    logger.info("copying running configuration to startup configuration")
    with optional_progress_bar(
        "Copying running config to startup config...",
        "Copied running config to startup config",
        "Failed to copy running config to startup config",
        display=progress_bar,
    ):
        config_command_index = 0
        commands = [
            {
                "command": "copy running-config startup-config",
                "success": r"\[OK\]",
                "questions": [{"prompt": r"Destination\sfilename\s\[startup\-config\]\?", "answer": ""}],
            }
        ]

        if enable_password:
            config_command_index = 1
            commands = with_enable_mode(commands, enable_password)

        output = run_list(cisco_switch, commands=commands, prompt_timeout=prompt_timeout, connection_timeout=connection_timeout)[
            config_command_index
        ]
        check_command_responses([(commands[config_command_index]["command"], output)])


def write_memory(
    cisco_switch: CiscoSwitch,
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 180,
    connection_timeout: Optional[int] = 10,
) -> None:
    """Copy the running configuration to startup-config.

This action is an alias for [copy_running_to_startup](#copy_running_to_startup).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __enable_password:__ the enable password if required (defaults to `null` assuming enable mode by default)

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `180`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.config.write_memory:
        progress_bar: true
      load:
        cisco_switch: cisco_switch
    ```
    """
    return copy_running_to_startup(
        cisco_switch,
        enable_password=enable_password,
        progress_bar=progress_bar,
        prompt_timeout=prompt_timeout,
        connection_timeout=connection_timeout,
    )
