"""
Interfaces

This module contains actions for viewing & managing interfaces on a Cisco IOS v15 switch.

See the [Overview](overview.md) page for information on how actions in this plugin work.
"""
import logging
import uuid

from typing import List, Union, Optional

from ppa import ui
from ssh_plugin_driver.client import run_list

from . import _images
from .records import CiscoSwitch, Interface
from ._generic_functions import (
    optional_progress_bar,
    parse_command_output,
    set_terminal_size,
    configure_object,
    check_command_responses,
    session,
)

logger = logging.getLogger(__name__)

TABLE_HEADER = ["Port", "Name", "Status", "Vlan", "Duplex", "Speed", "Type"]
TABLE_FIELDS = ["port", "name", "status", "vlan", "duplex", "speed", "type"]


@session
def get_all(
    cisco_switch: CiscoSwitch,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> List[Interface]:
    """Get all interfaces from the switch.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    A list of [Interface Dictionaries](../inputs_outputs/#interface).

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.get_all:
      load:
        cisco_switch: cisco_switch
      save: interfaces
    ```
    """
    with optional_progress_bar(
        "Getting all interfaces...", "Got all interfaces", "Failed to get all interfaces", display=progress_bar,
    ):
        command = "show interfaces status"
        output = run_list(
            cisco_switch,
            commands=set_terminal_size([{"command": command}]),
            prompt_timeout=prompt_timeout,
            connection_timeout=connection_timeout,
        )[-1]
        check_command_responses([(command, output)])
        return parse_command_output(command, output)


@session
def get_by_port(
    cisco_switch: CiscoSwitch,
    port: str,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> Interface:
    """Get an interface from the switch using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    A single [Interface Dictionary](../inputs_outputs/#interface).

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.get_by_port:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
      save: port_11
    ```
    """
    # Command "show interfaces status | include {port}" doesn't include
    # a table header, which is required by ntc_templates to parse the
    # results. Therefore this calls get_all and filters the returned list.
    with optional_progress_bar(
        f"Getting interface '{port}'...", f"Got interface '{port}'", f"Failed to get interface '{port}'", display=progress_bar,
    ):
        try:
            interface = list(
                filter(
                    lambda interface: interface["port"] == port,
                    get_all(cisco_switch, prompt_timeout=prompt_timeout, connection_timeout=connection_timeout),
                )
            )[0]
        except IndexError:
            raise ValueError(f"No interface found with port '{port}'")

    return interface


def _get_attribute(
    cisco_switch: CiscoSwitch,
    port: str,
    attribute: str,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> str:
    try:
        return get_by_port(cisco_switch, port, prompt_timeout=prompt_timeout, connection_timeout=connection_timeout)[attribute]
    except TypeError:
        raise ValueError(f"No interface found with port '{port}'.")
    except KeyError:
        raise ValueError(f"Interface with port '{port}' has no '{attribute}' attribute.")


@session
def get_speed(
    cisco_switch: CiscoSwitch,
    port: str,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> str:
    """Get the speed property of a switch interface.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    The interface speed as a string.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.get_speed:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
      save: port_11_speed
    ```
    """
    with optional_progress_bar(
        "Getting interface speed...", "Got interface speed", "Failed to get interface speed", display=progress_bar,
    ):
        return _get_attribute(cisco_switch, port, "speed", prompt_timeout=prompt_timeout, connection_timeout=connection_timeout)


@session
def get_duplex(
    cisco_switch: CiscoSwitch,
    port: str,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> str:
    """Get the duplex property of a switch interface.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    The interface duplex as a string.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.get_duplex:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
      save: port_11_duplex
    ```
    """
    with optional_progress_bar(
        "Getting interface duplex...", "Got interface duplex", "Failed to get interface duplex", display=progress_bar,
    ):
        return _get_attribute(cisco_switch, port, "duplex", prompt_timeout=prompt_timeout, connection_timeout=connection_timeout)


@session
def get_name(
    cisco_switch: CiscoSwitch,
    port: str,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> str:
    """Get the name of a switch interface.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    The interface name as a string.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.get_name:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
      save: port_11_name
    ```
    """
    with optional_progress_bar(
        "Getting interface name...", "Got interface name", "Failed to get interface name", display=progress_bar,
    ):
        return _get_attribute(cisco_switch, port, "name", prompt_timeout=prompt_timeout, connection_timeout=connection_timeout)


def _get_capabilities(
    cisco_switch: CiscoSwitch,
    port: str,
    attribute: str,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> List[str]:
    command = f"show interface {port} capabilities | include {attribute}"
    output = run_list(
        cisco_switch,
        commands=set_terminal_size([{"command": command}]),
        prompt_timeout=prompt_timeout,
        connection_timeout=connection_timeout,
    )[-1]
    check_command_responses([(command, output)])
    return [i.strip() for i in output.replace(f"{attribute}:", "").split(",")]


@session
def capable_speeds(
    cisco_switch: CiscoSwitch,
    port: str,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> List[str]:
    """Get the speeds an interface can support using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    A list of supported speeds as strings.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.capable_speeds:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
      save: port_11_speeds
    ```
    """
    # switch-001#show interface Gi0/23 capabilities | include Speed
    # Speed:                 10,100,1000,10000,auto
    with optional_progress_bar(
        "Getting valid interface speeds...",
        "Got valid interface speeds",
        "Failed to get valid interface speeds",
        display=progress_bar,
    ):
        return _get_capabilities(
            cisco_switch, port, "Speed", prompt_timeout=prompt_timeout, connection_timeout=connection_timeout
        )


@session
def capable_duplexes(
    cisco_switch: CiscoSwitch,
    port: str,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> List[str]:
    """Get the duplexes an interface can support using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    A list of supported duplexes as strings.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.capable_duplexes:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
      save: port_11_duplexes
    ```
    """
    # switch-001#show interface Gi0/23 capabilities | include Duplex
    # Duplex:                half,full,auto
    with optional_progress_bar(
        "Getting valid interface duplexes...",
        "Got valid interface duplexes",
        "Failed to get valid interface duplexes",
        display=progress_bar,
    ):
        return _get_capabilities(
            cisco_switch, port, "Duplex", prompt_timeout=prompt_timeout, connection_timeout=connection_timeout
        )


def _set_attribute(
    cisco_switch: CiscoSwitch,
    port: str,
    attribute: str,
    value: Union[str, int],
    enable_password: Optional[str] = None,
    error_messages: Optional[List[str]] = None,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> None:
    logger.info(f"setting {attribute} of interface port '{port}' to '{value}'")
    commands = set_terminal_size(configure_object("interface", port, [{"command": f"{attribute} {value}"}], enable_password))
    output_list = run_list(cisco_switch, commands=commands, prompt_timeout=prompt_timeout, connection_timeout=connection_timeout)

    # Check for known errors in the output of all commands
    check_command_responses(list(zip([c["command"] for c in commands], output_list)), error_messages=error_messages)


def _reset_attribute(
    cisco_switch: CiscoSwitch,
    port: str,
    attribute: str,
    enable_password: Optional[str] = None,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> None:
    logger.info(f"resetting {attribute} of interface port '{port}'")
    commands = set_terminal_size(configure_object("interface", port, [{"command": f"no {attribute}"}], enable_password))
    output_list = run_list(cisco_switch, commands=commands, prompt_timeout=prompt_timeout, connection_timeout=connection_timeout)

    # Check for known errors in the output of all commands
    check_command_responses(list(zip([c["command"] for c in commands], output_list)))


@session
def set_speed(
    cisco_switch: CiscoSwitch,
    port: str,
    speed: Union[str, None],
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> None:
    """Set the speed of an interface on the switch using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __speed:__ the new interface speed or `null` to set the default value

    - __enable_password:__ an [enable password](overview.md#enable-mode) if required

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.set_speed:
        port: Gi0/11
        speed: 100
      load:
        cisco_switch: cisco_switch
    ```
    """
    attribute = "speed"
    with optional_progress_bar(
        f"Setting interface speed to '{speed}'...",
        f"Interface speed has been set to '{speed}'",
        f"Failed to set interface speed to '{speed}'",
        display=progress_bar,
    ):
        _reset_attribute(
            cisco_switch,
            port,
            attribute,
            enable_password=enable_password,
            prompt_timeout=prompt_timeout,
            connection_timeout=connection_timeout,
        )
        if speed:
            return _set_attribute(
                cisco_switch,
                port,
                attribute,
                speed,
                enable_password=enable_password,
                error_messages=["Cannot change speed to 1000Mbps when in half duplex"],
                prompt_timeout=prompt_timeout,
                connection_timeout=connection_timeout,
            )


@session
def set_duplex(
    cisco_switch: CiscoSwitch,
    port: str,
    duplex: Union[str, None],
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> None:
    """Set the duplex of an interface on the switch using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __duplex:__ the new interface duplex or `null` to set the default value

    - __enable_password:__ an [enable password](overview.md#enable-mode) if required

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.set_duplex:
        port: Gi0/11
        duplex: full
      load:
        cisco_switch: cisco_switch
    ```
    """
    attribute = "duplex"
    with optional_progress_bar(
        f"Setting interface duplex to '{duplex}'...",
        f"Interface duplex has been set to '{duplex}'",
        f"Failed to set interface duplex to '{duplex}'",
        display=progress_bar,
    ):
        _reset_attribute(
            cisco_switch,
            port,
            attribute,
            enable_password=enable_password,
            prompt_timeout=prompt_timeout,
            connection_timeout=connection_timeout,
        )
        if duplex:
            _set_attribute(
                cisco_switch,
                port,
                attribute,
                duplex,
                enable_password=enable_password,
                error_messages=["Gigabit port is restricted to full duplex"],
                prompt_timeout=prompt_timeout,
                connection_timeout=connection_timeout,
            )


@session
def set_name(
    cisco_switch: CiscoSwitch,
    port: str,
    name: Union[str, None],
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> None:
    """Set the name of an interface on the switch using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __name:__ the new interface name or `null` to set an empty name

    - __enable_password:__ an [enable password](overview.md#enable-mode) if required

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.set_name:
        port: Gi0/11
        name: New Interface Name
      load:
        cisco_switch: cisco_switch
    ```
    """
    attribute = "description"
    with optional_progress_bar(
        f"Setting interface name to '{name}'...",
        f"Interface name has been set to '{name}'",
        f"Failed to set interface name to '{name}'",
        display=progress_bar,
    ):
        _reset_attribute(
            cisco_switch,
            port,
            attribute,
            enable_password=enable_password,
            prompt_timeout=prompt_timeout,
            connection_timeout=connection_timeout,
        )
        if name:
            _set_attribute(
                cisco_switch,
                port,
                attribute,
                name,
                enable_password=enable_password,
                prompt_timeout=prompt_timeout,
                connection_timeout=connection_timeout,
            )


@session
def set_default_speed(
    cisco_switch: CiscoSwitch,
    port: str,
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
):
    """Set the speed of an interface on the switch to its default value using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __enable_password:__ an [enable password](overview.md#enable-mode) if required

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.set_default_speed:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
    ```
    """
    with optional_progress_bar(
        "Setting interface speed to default...",
        "Interface speed has been set to default",
        "Failed to set interface speed to default",
        display=progress_bar,
    ):
        return _reset_attribute(
            cisco_switch,
            port,
            "speed",
            enable_password=enable_password,
            prompt_timeout=prompt_timeout,
            connection_timeout=connection_timeout,
        )


@session
def set_default_duplex(
    cisco_switch: CiscoSwitch,
    port: str,
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
):
    """Set the duplex of an interface on the switch to its default value using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __enable_password:__ an [enable password](overview.md#enable-mode) if required

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.set_default_duplex:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
    ```
    """
    with optional_progress_bar(
        "Setting interface duplex to default...",
        "Interface duplex has been set to default",
        "Failed to set interface duplex to default",
        display=progress_bar,
    ):
        return _reset_attribute(
            cisco_switch,
            port,
            "duplex",
            enable_password=enable_password,
            prompt_timeout=prompt_timeout,
            connection_timeout=connection_timeout,
        )


@session
def clear_access_vlan(
    cisco_switch: CiscoSwitch,
    port: str,
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
):
    """Remove the access VLAN of an interface on the switch using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __enable_password:__ an [enable password](overview.md#enable-mode) if required

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.clear_access_vlan:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
    ```
    """
    with optional_progress_bar(
        "Clearing interface access VLAN...",
        "Interface access VLAN has been cleared",
        "Failed to clear interface access VLAN",
        display=progress_bar,
    ):
        _reset_attribute(
            cisco_switch,
            port,
            "switchport access vlan",
            enable_password=enable_password,
            prompt_timeout=prompt_timeout,
            connection_timeout=connection_timeout,
        )


@session
def clear_name(
    cisco_switch: CiscoSwitch,
    port: str,
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
):
    """Clear the name of an interface on the switch using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __enable_password:__ an [enable password](overview.md#enable-mode) if required

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.clear_name:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
    ```
    """
    with optional_progress_bar(
        "Clearing interface name...", "Interface name has been cleared", "Failed to clear interface name", display=progress_bar,
    ):
        _reset_attribute(
            cisco_switch,
            port,
            "description",
            enable_password=enable_password,
            prompt_timeout=prompt_timeout,
            connection_timeout=connection_timeout,
        )


@session
def enable(
    cisco_switch: CiscoSwitch,
    port: str,
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> None:
    """Enable an interface on the switch using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __enable_password:__ an [enable password](overview.md#enable-mode) if required

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.enable:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
    ```
    """
    logger.info(f"enabling interface port '{port}'")
    with optional_progress_bar(
        "Enabling interface...", "Interface has been enabled", "Failed to enable interface", display=progress_bar,
    ):
        commands = set_terminal_size(configure_object("interface", port, [{"command": "no shutdown"}], enable_password))
        output_list = run_list(
            cisco_switch, commands=commands, prompt_timeout=prompt_timeout, connection_timeout=connection_timeout
        )
        # Check for known errors in the output of all commands
        check_command_responses(list(zip([c["command"] for c in commands], output_list)))


@session
def disable(
    cisco_switch: CiscoSwitch,
    port: str,
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> None:
    """Disable an interface on the switch using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __enable_password:__ an [enable password](overview.md#enable-mode) if required

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.disable:
        port: Gi0/11
      load:
        cisco_switch: cisco_switch
    ```
    """
    logger.info(f"disabling interface port '{port}'")
    with optional_progress_bar(
        "Disabling interface...", "Interface has been disabled", "Failed to disable interface", display=progress_bar,
    ):
        commands = set_terminal_size(configure_object("interface", port, [{"command": "shutdown"}], enable_password))
        output_list = run_list(
            cisco_switch, commands=commands, prompt_timeout=prompt_timeout, connection_timeout=connection_timeout
        )
        # Check for known errors in the output of all commands
        check_command_responses(list(zip([c["command"] for c in commands], output_list)))


@session
def set_access_vlan(
    cisco_switch: CiscoSwitch,
    port: str,
    vlan_id: Union[str, None],
    enable_password: Optional[str] = None,
    progress_bar: Optional[bool] = False,
    prompt_timeout: Optional[int] = 5,
    connection_timeout: Optional[int] = 10,
) -> None:
    """Set the access VLAN for an interface on the switch using its port name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __cisco_switch:__ a [CiscoSwitch](../inputs_outputs/#ciscoswitch) dictionary

    - __port:__ the interface port

    - __vlan_id:__ the ID of the new access VLAN or `null` to set a blank VLAN

    - __enable_password:__ an [enable password](overview.md#enable-mode) if required

    - __progress_bar:__ set to `true` to display a progress bar when performing the operation (defaults to `false`)

    - __prompt_timeout:__ the number of seconds to wait for commands to complete (defaults to `5`)

    - __connection_timeout:__ the number of seconds to wait for the connection to establish (defaults to `10`)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.set_access_vlan:
        port: Gi0/11
        vlan_id: 200
      load:
        cisco_switch: cisco_switch
    ```
    """
    attribute = "switchport access vlan"
    with optional_progress_bar(
        f"Setting access VLAN to '{vlan_id}'...",
        f"Access VLAN has been set to '{vlan_id}'",
        f"Failed to set access VLAN to '{vlan_id}'",
        display=progress_bar,
    ):
        _reset_attribute(
            cisco_switch,
            port,
            attribute,
            enable_password=enable_password,
            prompt_timeout=prompt_timeout,
            connection_timeout=connection_timeout,
        )
        if vlan_id:
            _set_attribute(
                cisco_switch,
                port,
                attribute,
                vlan_id,
                enable_password=enable_password,
                prompt_timeout=prompt_timeout,
                connection_timeout=connection_timeout,
            )


def input_table(
    text: str,
    interface_list: List[Interface],
    exclude_ports: Optional[List[str]] = None,
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
) -> List[Interface]:
    """Display a list of interfaces in a table, and allow the task operator to make a selection.

The table will have the following columns:

- Port
- Name
- Status
- VLAN
- Duplex
- Speed
- Type

!!! question "Minimum Plugin Version: 1.0.0"


??? info "Input"
    - __text:__ the title of the table

    - __interface_list:__ a list of [Interface Dictionaries](../inputs_outputs/#interface)

    - __exclude_ports:__ a list of interface ports to exclude from the table (defaults to `null`)

    - __minimum:__ the minimum number of acceptable selections (defaults to `null`)

    - __maximum:__ the maximum number of acceptable selections (defaults to `null`)

??? success "Output"
    A list of [Interface Dictionaries](../inputs_outputs/#interface).


??? example
    - Getting all interfaces from the switch and saving them as variable `all_interfaces`.

    - Displaying all interfaces other than the uplink, and saving the selection as `chosen_interfaces`.

    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.get_all:
      load:
        cisco_switch: cisco_switch
      save: all_interfaces

    - cisco_ios_switch.interfaces.input_table:
        text: Please Select an Interface
        exclude_ports:
          - Gi0/1
        minimum: 1
        maximum: 1
      load:
        interface_list: all_interfaces
      save: selected_interface
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())

    interface_list = (
        [interface for interface in interface_list if interface["port"] not in exclude_ports] if exclude_ports else interface_list
    )

    while True:
        indices = ui.input_table(
            text,
            header=TABLE_HEADER,
            data=[[interface.get(field) or "" for field in TABLE_FIELDS] for interface in interface_list],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum:
            if count < minimum:
                ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
                continue
        if maximum:
            if count > maximum:
                ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
                continue
        break

    return [interface_list[i] for i in indices]


def output_table(text: str, interface_list: List[Interface], exclude_ports: Optional[List[str]] = None) -> None:
    """Display a list of interfaces in a table.

The table will have the following columns:

- Port
- Name
- Status
- VLAN
- Duplex
- Speed
- Type

!!! question "Minimum Plugin Version: 1.0.0"


??? info "Input"
    - __text:__ the title of the table

    - __interface_list:__ a list of [Interface Dictionaries](../inputs_outputs/#interface)

    - __exclude_ports:__ a list of interface ports to exclude from the table (defaults to `null`)


??? success "Output"
    Nothing is outputted by this action.


??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.get_all:
      load:
        cisco_switch: cisco_switch
      save: all_interfaces

    - cisco_ios_switch.interfaces.output_table:
        text: All Interfaces
      load:
        interface_list: all_interfaces
    ```
    """
    interface_list = (
        [interface for interface in interface_list if interface["port"] not in exclude_ports] if exclude_ports else interface_list
    )
    ui.output_table(
        text, header=TABLE_HEADER, data=[[interface.get(field) or "" for field in TABLE_FIELDS] for interface in interface_list],
    )


def ethernet_status_icon(interface: Interface) -> str:
    """Generate an icon showing the interface port status as an HTML _img_ tag.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    - __interface:__ an [Interface Dictionary](../inputs_outputs/#interface)

??? success "Output"
    A snippet of HTML that can be used in markdown.

??? note "Screenshot"

    <img src="../../../images/screenshots/cisco/ethernet-icon.png" alt="ethernet-icon" width="140"/>

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.get_by_port:
        port: Gi0/10
      load:
        cisco_switch: cisco_switch
      save: interface_10

    - cisco_ios_switch.interfaces.ethernet_status_icon:
      load:
        interface: interface_10
      save: interface_10_icon

    - ppa.ui.output_markdown:
        doc: >
          #### Interface Gi0/10 Ethernet Icon

          {{ interface_10_icon }}
    ```
    """  # noqa

    def _format_html(interface, png_base_64):
        return f"<img src='data:image/png;base64,{png_base_64}' alt='{interface['port']}_status' style='float:left; margin-right: 40px; width: 145px' />"

    status_map = {"disabled": _images.ethernet_disabled, "notconnect": _images.ethernet_unplugged}
    speed_map = {
        "10": _images.ethernet_10,
        "100": _images.ethernet_100,
        "1000": _images.ethernet_1000,
    }
    try:
        # If it's not connected or disabled return an image from status_map.
        return _format_html(interface, status_map[interface["status"]])
    except KeyError:
        try:
            # Otherwise it should be connected with a valid speed. Speeds
            # are displayed either as 'a-[speed]', or just '[speed]', so
            # replace 'a-' from the value if it exists before the look-up.
            return _format_html(interface, speed_map[interface["speed"].replace("a-", "")])
        except KeyError:
            raise ValueError(
                f"Could not find ethernet port icon for speed '{interface['speed']}' and status '{interface['status']}'."
            )


def ethernet_summary(interface: Interface) -> str:
    """Generate an HTML interface summary with an icon showing the interface port status.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    - __interface:__ an [Interface Dictionary](../inputs_outputs/#interface)

??? success "Output"
    A snippet of HTML that can be used in markdown.

??? note "Screenshot"

    <img src="../../../images/screenshots/cisco/ethernet-summary.png" alt="ethernet-summary" width="300"/>

??? example
    ```YAML linenums="1"
    - cisco_ios_switch.interfaces.get_by_port:
        port: Gi0/10
      load:
        cisco_switch: cisco_switch
      save: interface_10

    - cisco_ios_switch.interfaces.ethernet_summary:
      load:
        interface: interface_10
      save: interface_10_summary

    - ppa.ui.output_markdown:
        doc: >
          #### Interface Gi0/10 Summary

          {{ interface_10_summary }}
    ```
    """  # noqa
    return f"""
<p>
    {ethernet_status_icon(interface)}
    <strong>Port: </strong>{interface['port']}<br>
    <strong>Speed: </strong>{interface['speed']}<br>
    <strong>Duplex: </strong>{interface['duplex']}<br>
    <strong>Status: </strong>{interface['status']}<br>
    <strong>Name: </strong>{interface['name']}
</p>
"""
